@@ -4,23 +4,34 @@
    result.  If STRING is not NULL, the caller must ensure enough space is
    available to store the result.
 
-Copyright 1991, 1993, 1994, 1996, 2000, 2001, 2002, 2005 Free Software
+Copyright 1991, 1993, 1994, 1996, 2000-2002, 2005, 2012 Free Software
 Foundation, Inc.
 
 This file is part of the GNU MP Library.
 
 The GNU MP Library is free software; you can redistribute it and/or modify
-it under the terms of the GNU Lesser General Public License as published by
-the Free Software Foundation; either version 3 of the License, or (at your
-option) any later version.
+it under the terms of either:
+
+  * the GNU Lesser General Public License as published by the Free
+    Software Foundation; either version 3 of the License, or (at your
+    option) any later version.
+
+or
+
+  * the GNU General Public License as published by the Free Software
+    Foundation; either version 2 of the License, or (at your option) any
+    later version.
+
+or both in parallel, as here.
 
 The GNU MP Library is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
-License for more details.
+or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
+for more details.
 
-You should have received a copy of the GNU Lesser General Public License
-along with the GNU MP Library.  If not, see http://www.gnu.org/licenses/.  */
+You should have received copies of the GNU General Public License and the
+GNU Lesser General Public License along with the GNU MP Library.  If not,
+see https://www.gnu.org/licenses/.  */
 
 #include <string.h> /* for strlen */
 #include "gmp.h"
@@ -31,19 +42,18 @@ char *
 mpz_get_str (char *res_str, int base, mpz_srcptr x)
 {
   mp_ptr xp;
-  mp_size_t x_size = x->_mp_size;
-  char *str;
+  mp_size_t x_size = SIZ (x);
   char *return_str;
   size_t str_size;
   size_t alloc_size = 0;
-  char *num_to_text;
+  const char *num_to_text;
   int i;
   TMP_DECL;
 
   if (base >= 0)
     {
       num_to_text = "0123456789abcdefghijklmnopqrstuvwxyz";
-      if (base == 0)
+      if (base <= 1)
 	base = 10;
       else if (base > 36)
 	{
@@ -55,6 +65,10 @@ mpz_get_str (char *res_str, int base, mpz_srcptr x)
   else
     {
       base = -base;
+      if (base <= 1)
+	base = 10;
+      else if (base > 36)
+	return NULL;
       num_to_text = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
     }
 
@@ -76,29 +90,19 @@ mpz_get_str (char *res_str, int base, mpz_srcptr x)
 
   /* mpn_get_str clobbers its input on non power-of-2 bases */
   TMP_MARK;
-  xp = x->_mp_d;
+  xp = PTR (x);
   if (! POW2_P (base))
     {
-      xp = TMP_ALLOC_LIMBS (x_size + 1);  /* +1 in case x_size==0 */
-      MPN_COPY (xp, x->_mp_d, x_size);
+      xp = TMP_ALLOC_LIMBS (x_size | 1);  /* |1 in case x_size==0 */
+      MPN_COPY (xp, PTR (x), x_size);
     }
 
   str_size = mpn_get_str ((unsigned char *) res_str, base, xp, x_size);
   ASSERT (alloc_size == 0 || str_size <= alloc_size - (SIZ(x) < 0));
 
-  /* might have a leading zero, skip it */
-  str = res_str;
-  if (*res_str == 0 && str_size != 1)
-    {
-      str_size--;
-      str++;
-      ASSERT (*str != 0);  /* at most one leading zero */
-    }
-
-  /* Convert result to printable chars, and move down if there was a leading
-     zero.  */
+  /* Convert result to printable chars.  */
   for (i = 0; i < str_size; i++)
-    res_str[i] = num_to_text[str[i]];
+    res_str[i] = num_to_text[(int) res_str[i]];
   res_str[str_size] = 0;
 
   TMP_FREE;
@@ -109,7 +113,7 @@ mpz_get_str (char *res_str, int base, mpz_srcptr x)
       size_t  actual_size = str_size + 1 + (res_str - return_str);
       ASSERT (actual_size == strlen (return_str) + 1);
       __GMP_REALLOCATE_FUNC_MAYBE_TYPE (return_str, alloc_size, actual_size,
-                                        char);
+					char);
     }
   return return_str;
 }
